﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Text;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;
#if SPYRO_LINQ2
using Spyro.Linq2;
#else
using System.Linq;
#endif

namespace $rootnamespace$.Xml
{
    /// <summary>
    /// Proporciona métodos para el manejo de Xml
    /// </summary>
    public static partial class UtilidadesXml
    {
        public static GeneracionXml Generar<T>(XmlSerializerNamespaces p_namespaces, T p_objeto, string[] p_esquemas, string p_destino)
        {
            var v_resultado = GenerarYValidar(p_namespaces, p_objeto, p_esquemas);

            if (v_resultado.Advertencias != null && v_resultado.Advertencias.Length > 0)
                Trace.WriteLine(string.Format("Advertencias:\n{0}", v_resultado.Advertencias.Aggregate((p_primero, p_segundo) => p_primero + "\n" + p_segundo)));

            if (v_resultado.Errores != null && v_resultado.Errores.Length > 0)
            {
                Trace.WriteLine(string.Format("Errores:\n{0}", v_resultado.Errores.Aggregate((p_primero, p_segundo) => p_primero + "\n" + p_segundo)));
                //throw new Exception(v_resultado.Errores[0]);
            }

            if (!string.IsNullOrEmpty(p_destino) && (v_resultado.Errores == null || v_resultado.Errores.Length == 0))
                File.WriteAllText(p_destino, v_resultado.Xml);

            return v_resultado;
        }

        internal sealed class Utf8StringWriter : StringWriter
        {
            public override Encoding Encoding { get { return Encoding.UTF8; } }
        }

        //[AttributeUsage(AttributeTargets.Property)]
        //public class DecimalFormatterAttribute : Attribute
        //{
        //    public DecimalFormatterAttribute(string formatString)
        //    {
        //        Format = formatString;
        //    }
        //
        //    public string Format { get; private set; }
        //}

        private static GeneracionXml GenerarYValidar<T>(XmlSerializerNamespaces p_namespaces, T p_objeto, string[] p_esquemas)
        {
            var v_serializador = new XmlSerializer(typeof(T));
            var v_settingsSerializador = new XmlWriterSettings();
            v_settingsSerializador.Encoding = Encoding.UTF8;
            //v_settingsSerializador.NewLineOnAttributes = true;
            v_settingsSerializador.Indent = true;

            using (var v_stringWriter = new Utf8StringWriter())
            using (var v_writerXml = XmlWriter.Create(v_stringWriter, v_settingsSerializador))
            //using (var v_writerXml = XmlWriter.Create(@"u:\temp\rbm\xx.xml", v_settingsSerializador)) 
            {
                v_serializador.Serialize(v_writerXml, p_objeto, p_namespaces);
                var v_xml = v_stringWriter.ToString();

                string[] v_advertencias, v_errores;
                if (p_esquemas != null)
                {
                    var v_ensamblado = Assembly.GetExecutingAssembly();
                    var v_streamsEsquemas = new Stream[p_esquemas.Length];
                    try
                    {
                        for (var i = 0; i < p_esquemas.Length; i++)
                        {
                            v_streamsEsquemas[i] = v_ensamblado.GetManifestResourceStream(p_esquemas[i]);
                        }
                        Validar(v_xml, v_streamsEsquemas, out v_advertencias, out v_errores);
                    }
                    finally
                    {
                        for (var i = 0; i < p_esquemas.Length; i++)
                        {
                            if (v_streamsEsquemas[i] != null)
                            {
                                v_streamsEsquemas[i].Dispose();
                                v_streamsEsquemas[i] = null;
                            }

                        }
                    }
                }
                else
                    v_advertencias = v_errores = new string[0];

                var v_resultado = new GeneracionXml(v_xml, v_advertencias, v_errores);
                return v_resultado;
            }
        }

        public class GeneracionXml
        {
            public string Xml { get; private set; }
            public string[] Advertencias { get; private set; }
            public string[] Errores { get; private set; }

            public GeneracionXml(string p_xml, string[] p_advertencias, string[] p_errores)
            {
                this.Xml = p_xml;
                this.Advertencias = p_advertencias;
                this.Errores = p_errores;
            }
        }

        /// <summary>
        /// Valida el Xml y devuelve un par de arrays con las advertencias y errores encontrados. 
        /// </summary>
        public static void Validar(string p_xml, Stream[] p_esquemas, out string[] p_advertencias, out string[] p_errores)
        {
            for( var i = 0; i < p_esquemas.Length; i++) { 
}
            using (var v_stringReader = new StringReader(p_xml))
            {
                Validar(v_stringReader, p_esquemas, out p_advertencias, out p_errores);
            }
        }


        /// <summary>
        /// Valida el Xml y devuelve un par de arrays con las advertencias y errores encontrados. 
        /// </summary>
        public static void Validar(TextReader p_xml, Stream[] p_esquemas, out string[] p_advertencias, out string[] p_errores)
        {
            var v_readersEsquema = new XmlReader[p_esquemas.Length];
            try 
            {
                // Set the validation settings.
                var v_settingsEsquema = new XmlReaderSettings();
                v_settingsEsquema.ValidationType = ValidationType.Schema;
                v_settingsEsquema.ValidationFlags |= XmlSchemaValidationFlags.ProcessInlineSchema;
                v_settingsEsquema.ValidationFlags |= XmlSchemaValidationFlags.ProcessSchemaLocation;
                v_settingsEsquema.ValidationFlags |= XmlSchemaValidationFlags.ReportValidationWarnings;

                var v_advertencias = new List<string>();
                var v_errores = new List<string>();

                v_settingsEsquema.ValidationEventHandler += GenerarCallbackValidacion(v_advertencias, v_errores);
                for (var i = 0; i < p_esquemas.Length; i++)
                {
                    v_readersEsquema[i] = XmlReader.Create(p_esquemas[i]);
                    v_settingsEsquema.Schemas.Add(null, v_readersEsquema[i]);
                }


                // Create the XmlReader object.
                using (var v_readerXml = XmlReader.Create(p_xml, v_settingsEsquema))
                {
                    // Parse the file. 
                    try
                    {
                        while (v_readerXml.Read()) ;
                    }
                    catch
                    {
                    }
                }

                p_advertencias = v_advertencias.ToArray();
                p_errores = v_errores.ToArray();
            }
            finally
            {
                for (var i = 0; i < p_esquemas.Length; i++)
                {
                    ((IDisposable)v_readersEsquema[i]).Dispose();
                    v_readersEsquema[i] = null;
                }
            }
        }

        private static ValidationEventHandler GenerarCallbackValidacion(List<string> p_advertencias, List<string> p_errores)
        {
            ValidationEventHandler v_callBack = (p_sender, p_args) =>
            {
                if (p_args.Severity == XmlSeverityType.Warning)
                    p_advertencias.Add(p_args.Message);
                else
                    p_errores.Add(p_args.Message);
            };

            return v_callBack;
        }
    }
}